import inspect
from collections import defaultdict

class PipelineForwardError(Exception):
    def __init__(self, debug_memory, func, used_args):
        message = self.__error_message(debug_memory, func, used_args)
        super(PipelineForwardError, self).__init__(message)


    def __error_message(self, debug_memory, func, used_args):
        return "\n".join([
            f"Returned value of '{debug_memory.last.__name__}' does not match the arguement of '{func.__name__}'.",
            f"  '{debug_memory.last.__name__}' returned: {used_args}",
            f"  '{func.__name__}' consumes: {format_args(func)}",
            f"  {func_details(debug_memory.last)}",
            f"  {func_details(func)}"
        ])


class PipelineCallError(Exception):
    def __init__(self, debug_memory, func, used_args):
        message = self.__error_message(debug_memory, func, used_args)
        super(PipelineCallError, self).__init__(message)

    def __error_message(self, debug_memory, func, used_args):
        return "\n".join([
            f"Given arguments do not match '{debug_memory.first.__name__}' arguments.",
            f"  Given arguments: {used_args}",
            f"  '{debug_memory.first.__name__}' consumes: {format_args(debug_memory.first)}",
            f"  {func_details(debug_memory.first)}"
        ])


class MergeError(Exception):
    def __init__(self, func, used_args):
        message = self.__error_message(func, used_args)
        super(MergeError, self).__init__(message)

    def __error_message(self, func, used_args):
        return "\n".join([
            f"Given arguments do not match '{func.__name__}' arguments",
            f"  Given arguments: {used_args}",
            f"  '{func.__name__}' consumes: {format_args(func)}",
            f"  {func_details(func)}"
        ])


def format_args(func):
    argspec = inspect.getfullargspec(func)
    annotations = defaultdict(lambda: None)
    for k, v in argspec.annotations.items():
        annotations[k] = v

    def _annotate(annotation):
        if annotation is None:
            return ""
        return f": {annotation}"

    args = ', '.join([f'{x}{_annotate(annotations[x])}' for x in argspec.args])
    return f'({args})'


def func_details(func):
    return f"'{func.__name__}' from File \"{func.__code__.co_filename}\", line {func.__code__.co_firstlineno}"




from pypely.memory._impl import get_memory, set_memory, PipelineMemory


class PipelineMemoryContext:
    previous_memory: PipelineMemory = None

    def __enter__(self):
        self.previous_memory = get_memory()
        set_memory(PipelineMemory())

    def __exit__(self, type, value, traceback):
        set_memory(self.previous_memory)



from pypely.memory._impl import get_memory
from pypely.memory.errors import MemoryIngestNotAllowedError
from typing import Callable, TypeVar, Optional, Union

T = TypeVar("T")

Memorizable = Callable[[Callable[..., T]], Callable[..., T]]


def memorizable(func: Optional[Callable[..., T]]=None, allow_ingest: Optional[bool]=True) -> Union[Memorizable, Callable[..., T]]:
    class Memorizable:
        def __init__(self, func):
            self.func = func
            self.attributes_after = []
            self.attributes_before = []

        def __rshift__(self, other):
            return _add_to_memory(self.func, other)

        def __lshift__(self, other):
            self.__check_ingest()
            self.attributes_after.append(other)
            return self

        def __rrshift__(self, other):
            self.__check_ingest()
            self.attributes_before.append(other)
            return self

        def __call__(self, *args):
            memory = get_memory()
            memory_attributes_before = [memory.get(attr) for attr in self.attributes_before]
            memory_attributes_after = [memory.get(attr) for attr in self.attributes_after]
            self.__reset()

            return self.func(*memory_attributes_before, *args, *memory_attributes_after)

        def __reset(self):
            self.attributes_after = []
            self.attributes_before = []

        def __check_ingest(self):
            if not allow_ingest:
                raise MemoryIngestNotAllowedError(f"Memory ingest is not allowed for func: {self.func.__qualname__}")
    
    if func is None:
        return Memorizable
    else:
        return Memorizable(func)


def _add_to_memory(func: Callable[..., T], name: str, ) -> Callable[..., T]:
    def __inner(*args):
        result = func(*args)
        memory = get_memory()
        memory.add(name, result)

        return result
    return __inner



from collections import namedtuple
from typing import Callable
from pypely.core.errors import PipelineCallError, PipelineForwardError

DebugMemory = namedtuple('DebugMemory', ['combine', 'first', 'last'])

def debugable_reduce(debug_memory: DebugMemory, func2: Callable):
    def _combine(*args):
        _first = _try(debug_memory.combine, debug_memory, PipelineCallError)
        _second = _try(func2, debug_memory, PipelineForwardError)

        return _second(_first(*args))

    return DebugMemory(combine=_combine, last=func2, first=debug_memory.first)


def _try(func, debug_memory, error_cls):
    def wrapper(*args):
        try:
            return func(*args)
        except TypeError:
            raise error_cls(debug_memory, func, args)

    return wrapper



class MemoryAttributeExistsError(AttributeError):
    def __init__(self, message):
        super(AttributeError, self).__init__(message)


class MemoryAttributeNotFoundError(AttributeError):
    def __init__(self, message):
        super(AttributeError, self).__init__(message)


class MemoryIngestNotAllowedError(RuntimeError):
    def __init__(self, message):
        super(RuntimeError, self).__init__(message)


from typing import Optional, Any
from pypely.memory.errors import MemoryAttributeExistsError, MemoryAttributeNotFoundError

class PipelineMemory:
    def add(self, name: str, value: Any) -> None:
        if name in self.__dict__.keys():
            raise MemoryAttributeExistsError(f"The attribute {name} already exists")
        self.__setattr__(name, value)

    def get(self, name: str) -> Any:
        try:
            return self.__getattribute__(name)
        except AttributeError:
            raise MemoryAttributeNotFoundError(
                f"""Tried to access attribute {name} but attribute was not found. 
                Available Attributes: {list(self.__dict__.keys())}"""
            )


MEMORY: Optional[PipelineMemory] = None

def get_memory() -> PipelineMemory:
    global MEMORY
    if MEMORY is None:
        MEMORY = PipelineMemory()
    return MEMORY


def set_memory(memory: PipelineMemory) -> None:
    global MEMORY
    MEMORY = memory


from functools import reduce
from typing import Callable, TypeVar
from pypely.helpers import flatten
from pypely._types import PypelyTuple
from pypely.memory import memorizable
from pypely.memory._context import PipelineMemoryContext
from pypely.core._debug_helpers import debugable_reduce, DebugMemory
from pypely.core.errors import MergeError

T = TypeVar("T")


@memorizable
def pipeline(*funcs: Callable) -> Callable:
    initial = DebugMemory(combine=funcs[0], first=funcs[0], last=funcs[0])
    _pipeline = reduce(debugable_reduce, funcs[1:], initial).combine

    def _call(*args):
        with PipelineMemoryContext() as _:
            return _pipeline(*args)

    return _call


@memorizable(allow_ingest=False)
def fork(*funcs: Callable) -> Callable[..., PypelyTuple]:
    return lambda *x: PypelyTuple(*[func(*x) for func in funcs])


@memorizable(allow_ingest=False)
def to(obj: T, *set_fields: str) -> Callable[[PypelyTuple], T]:
    def _inner(vals: PypelyTuple) -> T:
        vals_flattened = flatten(vals)
        if not set_fields == ():
            assert len(vals_flattened) == len(set_fields)
            fields_named = {field_name: val for field_name, val in zip(set_fields, vals_flattened)}
            return obj(**fields_named)
        else:
            return obj(*vals_flattened)
    
    return _inner


@memorizable(allow_ingest=False)
def merge(func: Callable[..., T]) -> Callable[[PypelyTuple], T]:
    def _inner(branches):
        flat_branches = flatten(branches)
        try:
            return func(*flat_branches)
        except TypeError:
            raise MergeError(func, branches)

    return _inner


def identity(*x):
    if len(x) == 1:
        return x[0]
    return x
